#include "filemanagerservice.h"


Binder::Binder(QWebFrame *webFrame) :
    m_uuid(QUuid::createUuid()),
    m_webFrame(webFrame)
{
}


QString Binder::uuid()
{
    return m_uuid.toString();
}


void Binder::success(QString &ret)
{
    m_reply = ret;
    m_webFrame->evaluateJavaScript(QString("deosc.file_manager_service_success('%1')").arg(uuid()));
}


void Binder::error(QString &err)
{
    m_reply = err;
    m_webFrame->evaluateJavaScript(QString("deosc.file_manager_service_error('%1')").arg(uuid()));
}


QString Binder::reply()
{
    return m_reply;
}


Metadata::Metadata (QFileInfo info) : m_info(info) {}


void Metadata::appendChild(Metadata &metadata)
{
    m_children << metadata;
}


QDomElement Metadata::createDomElementFromFileInfo (QDomDocument &doc, const QFileInfo &info)
{
    QRegExp re("^Documents");

    QDomElement metadata = doc.createElement("metadata");

    QDomElement path = doc.createElement("path");
    QDomText pathTxt = doc.createTextNode(info.filePath().replace(re, ""));
    path.appendChild(pathTxt);
    metadata.appendChild(path);

    QDomElement name = doc.createElement("name");
    QDomText nameTxt = doc.createTextNode(info.fileName());
    name.appendChild(nameTxt);
    metadata.appendChild(name);

    QDomElement lastModified = doc.createElement("lastModified");
    QDomText lastModifiedTxt = doc.createTextNode(QString::number(info.lastModified().toMSecsSinceEpoch()));
    lastModified.appendChild(lastModifiedTxt);
    metadata.appendChild(lastModified);

    QDomElement lastRead = doc.createElement("lastRead");
    QDomText lastReadTxt = doc.createTextNode(QString::number(info.lastRead().toMSecsSinceEpoch()));
    lastRead.appendChild(lastReadTxt);
    metadata.appendChild(lastRead);

    QDomElement isdir = doc.createElement("isdir");
    metadata.appendChild(isdir);
    QDomText isdirTxt = doc.createTextNode(info.isDir() ? "True" : "False");
    isdir.appendChild(isdirTxt);

    return metadata;
}


QDomElement Metadata::element (QDomDocument &doc)
{
    QDomElement metadata = createDomElementFromFileInfo (doc, m_info);
    doc.appendChild(metadata);

    if (m_info.isDir()) {
        QDomElement children = doc.createElement("children");
        metadata.appendChild(children);

        foreach (Metadata m, m_children)
            children.appendChild(m.element(doc));
    }
    return metadata;
}


QDomDocument Metadata::document ()
{
    QDomDocument doc;
    element(doc);
    return doc;
}


QString Metadata::toString()
{
    QDomDocument doc;
    QDomElement e = element(doc);
    doc.appendChild(e);

    return doc.toString();
}


FileManagerService::FileManagerService(QObject *parent, Html5ApplicationViewer *html5ApplicationViewer) :
    QObject(parent),
    m_html5ApplicationViewer(html5ApplicationViewer),
    m_webView(html5ApplicationViewer->webView())
{
    m_webView->setAcceptTouchEvents(false);
    m_webView->setAcceptHoverEvents(true);
    m_html5ApplicationViewer->setAttribute(Qt::WA_AcceptTouchEvents, false);

    try {
        QDir dir = QDir::home();
        if (!dir.exists("DCaseWeaver")) {
            if (!dir.mkdir("DCaseWeaver")) {
                throw QString("mkdir failed: %1").arg(dir.filePath("DCaseWeaver"));
            }
        }

        if (dir.cd("DCaseWeaver")) {
            if (!dir.exists("Documents")) {
                if (!dir.mkdir("Documents")) {
                    throw QString("mkdir failed: %1").arg(dir.filePath("Documents"));
                }
            }
            if (!dir.exists(".localstorage")) {
                if (!dir.mkdir(".localstorage")) {
                    throw QString("mkdir failed: %1").arg(dir.filePath(".localstorage"));
                }
            }
        } else {
            throw QString("No such directory: %1").arg(dir.filePath("DCaseWeaver"));
        }

        m_webView->page()->settings()->setAttribute(QWebSettings::LocalStorageEnabled, true);
        m_webView->page()->settings()->setLocalStoragePath (dir.filePath(".localstorage"));

    } catch (QString err) {
        //qDebug() << err;
        QMessageBox::critical( 0, "Application name here",
                QString("An internal error occurred. Please ") +
                "call technical support at 123456789 and report\n"+
                "these numbers:\n\n" + err +
                "\n\n<Application> will now exit." );
        connect(m_html5ApplicationViewer, SIGNAL(show()), SLOT(quit()));
    }

    QDir::setCurrent(QDir::home().filePath("DCaseWeaver"));

    connect(m_webView->page()->mainFrame(),
            SIGNAL(javaScriptWindowObjectCleared()), SLOT(addToJavaScript()));
    connect(this, SIGNAL(signal_network_finished(QNetworkAccessManager *)),
            SLOT(network_finished(QNetworkAccessManager *)), Qt::QueuedConnection);
}


void FileManagerService::addToJavaScript()
{
    m_webView->page()->mainFrame()->addToJavaScriptWindowObject("IFileManagerService", this);
}


QString FileManagerService::bindService()
{
    //qDebug() << "bindService";
    Binder *binder = new Binder(m_webView->page()->mainFrame());
    m_binders.insert(binder->uuid(), binder);
    return binder->uuid();
}


void FileManagerService::unbindService(QString uuid)
{
    //qDebug() << "unbindService";
    Binder *binder = m_binders.value(uuid);
    delete binder;
    m_binders.remove(uuid);
}


void FileManagerService::execute(QString uuid, QString method, QStringList args)
{
    //qDebug() << "execute";
    Binder *binder = m_binders.value(uuid);

    if (method == "metadata") {
        metadata(binder, args);

    } else if (method == "parents") {
        parents(binder, args);

    } else if (method == "children") {
        children(binder, args);

    } else if (method == "get_file") {
        get_file(binder, args);

    } else if (method == "put_file") {
        put_file(binder, args);

    } else if (method == "post_file") {
        post_file(binder, args);

    } else if (method == "mkdir") {
        mkdir(binder, args);

    } else if (method == "remove") {
        remove(binder, args);

    } else if (method == "save_as_image") {
        save_as_image(binder, args);

    } else if (method == "network") {
        network(binder, args);
    }
}


QString FileManagerService::reply(QString uuid)
{
    //qDebug() << "reply: ";
    return m_binders.value(uuid)->reply();
}


void FileManagerService::metadata(Binder *binder, QStringList &args)
{
    //qDebug() << "metadata";

    try {
        QString path = args.takeFirst().prepend("Documents");

        QFileInfo info(path);
        if (!info.exists()) {
            throw QString("No such file or directory: %1").arg(info.absoluteFilePath());
        }

        Metadata metadata(info);

        if (info.isDir()) {
            QDir dir = QDir(info.filePath());
            foreach(QFileInfo i, dir.entryInfoList(QDir::Dirs|QDir::NoDotAndDotDot)) {
                Metadata m(i);
                metadata.appendChild(m);
            }
            foreach(QFileInfo i, dir.entryInfoList(QDir::Files)) {
                Metadata m(i);
                metadata.appendChild(m);
            }
        }

        QString ret = metadata.toString();
        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::parents(Binder *binder, QStringList &args)
{
    //qDebug() << "parents";
    try {
        QString path = args.takeFirst().prepend("Documents");

        QFileInfo info(path);
        if (!info.exists()) {
            throw QString("No such file or directory: %1").arg(info.absoluteFilePath());
        }
        QString parentPath = info.path();
        QFileInfo parent(parentPath);
        if (!parent.exists()) {
            throw QString("No such file or directory: %1").arg(parent.absoluteFilePath());
        }

        QDomDocument doc;
        QDomElement e = doc.createElement("parents");
        doc.appendChild(e);

        Metadata m(parent);
        e.appendChild(m.element(doc));

        QString ret = doc.toString();
        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::children(Binder *binder, QStringList &args)
{
    //qDebug() << "children";

    try {
        QString path = args.takeFirst().prepend("Documents");

        QFileInfo info(path);
        if (!info.exists()) {
            throw QString("No such directory: %1").arg(info.absoluteFilePath());
        }
        if (!info.isDir()) {
            throw QString("Not a directory: %1").arg(info.absoluteFilePath());
        }

        QDomDocument doc;
        QDomElement e = doc.createElement("children");
        doc.appendChild(e);

        QDir dir = QDir(info.filePath());
        foreach(QFileInfo i, dir.entryInfoList(QDir::Dirs|QDir::NoDotAndDotDot)) {
            Metadata m(i);
            e.appendChild(m.element(doc));
        }
        foreach(QFileInfo i, dir.entryInfoList(QDir::Files)) {
            Metadata m(i);
            e.appendChild(m.element(doc));
        }

        QString ret = doc.toString();
        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::get_file(Binder *binder, QStringList &args)
{
    //qDebug() << "get_file";

    try {
        QString path = args.takeFirst().prepend("Documents");
        QString type = args.takeFirst();

        QFileInfo info(path);
        if (!info.exists()) {
            throw QString("No such file: %1").arg(info.absoluteFilePath());
        }
        if (!info.isFile()) {
            throw QString("Not a file: %1").arg(info.absoluteFilePath());
        }
        QFile f(path);
        if (!f.open(QIODevice::ReadOnly)) {
            throw QString("Failed to open file: %1").arg(info.absoluteFilePath());
        }

        QString ret;
        if (type == "text") {
            ret.append(f.readAll());
        } else {
            ret.append(f.readAll().toBase64());
        }

        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::put_file(Binder *binder, QStringList &args)
{
    //qDebug() << "put_file";
    try {
        QString path = QString("Documents").append(args[0]);
        //qDebug() << "path: " << path;
        QString &type = args[1];
        QString &data = args[2];

        QFileInfo info(path);
        if (info.exists() && !info.isFile()) {
            throw QString("Not a file: %1").arg(info.absoluteFilePath());
        }
        QFile file(path);
        if (!file.open(QIODevice::WriteOnly)) {
            throw QString("Failed to open file: %1").arg(info.absoluteFilePath());
        }
        if (type == "text") {
            if (file.write(data.toUtf8()) == -1) {
                throw QString("Failed to write file: %1").arg(info.absoluteFilePath());
            }
        } else {
            if (file.write(QByteArray::fromBase64(data.toLocal8Bit())) == -1) {
                throw QString("Failed to write file: %1").arg(info.absoluteFilePath());
            }
        }

        info.refresh();
        Metadata metadata(info);
        QString ret = metadata.toString();

        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::post_file(Binder *binder, QStringList &args)
{
    //qDebug() << "post_file";

    try {
        QString path = QString("Documents").append(args[0]);
        QString &name = args[1];
        QString &type = args[2];
        QString &data = args[3];

        QDir dir(path);
        QString filePath = dir.filePath(name);
        QFile file(filePath);

        if (!dir.exists()) {
            throw QString("No such directory: %1").arg(dir.absolutePath());
        }
        if (file.exists()) {
            throw QString("File already exists: %1").arg(dir.absoluteFilePath(name));
        }
        if (!file.open(QIODevice::WriteOnly)) {
            throw QString("Failed to open file: %1").arg(dir.absoluteFilePath(name));
        }
        if (type == "text") {
            if (file.write(data.toUtf8()) == -1) {
                throw QString("Failed to write file: %1").arg(dir.absoluteFilePath(name));
            }
        } else {
            if (file.write(QByteArray::fromBase64(data.toLocal8Bit())) == -1) {
                throw QString("Failed to write file: %1").arg(dir.absoluteFilePath(name));
            }
        }

        QFileInfo info(filePath);
        Metadata metadata(info);
        QString ret = metadata.toString();

        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::mkdir(Binder *binder, QStringList &args)
{
    //qDebug() << "mkdir";
    try {
        QString path = QString("Documents").append(args[0]);
        QString &name = args[1];

        QFileInfo info(path);
        if (!info.isDir()) {
            throw QString("Not a directory: %1").arg(info.absoluteFilePath());
        }
        QDir dir(path);
        if (dir.exists(name)) {
            throw QString("The directory already exists: %1").arg(dir.absoluteFilePath(name));
        }
        if (!dir.mkdir(name)) {
            throw QString("Failed to mkdir: %1").arg(dir.absoluteFilePath(name));
        }

        QFileInfo newDirInfo(dir.filePath(name));
        Metadata metadata(newDirInfo);
        QString ret = metadata.toString();

        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::remove(Binder *binder, QStringList &args)
{
    //qDebug() << "remove";
    try {
        QString path = QString("Documents").append(args[0]);

        QFileInfo info(path);
        QString name = info.fileName();
        QDir parentDir = info.dir();

        Metadata metadata(info);

        if (info.isDir()) {
            QDir dir(path);
            dir.setFilter(QDir::AllEntries|QDir::NoDotAndDotDot);
            if (0 < dir.count()) {
                throw QString("The directory is not empty: %1").arg(info.absoluteFilePath());
            }
            if (!parentDir.rmdir(name)) {
                throw QString("Failed to remove directory: %1").arg(info.absoluteFilePath());
            }
        } else if (info.isFile()) {
            if (!parentDir.remove(name)) {
                throw QString("Failed to remove file: %1").arg(info.absoluteFilePath());
            }
        } else {
            throw QString("Failed to remove: %1").arg(info.absoluteFilePath());
        }

        QString ret = metadata.toString();

        binder->success(ret);

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::save_as_image(Binder *binder, QStringList &args)
{
    //qDebug() << "save_as_image";

    try {
        QString path = args.takeFirst().prepend("Documents");
        QString name = args.takeFirst();
        QStringList &images = args;
        QString basename = QFileInfo(name).baseName();

        QDir dir(path);
        if (!dir.exists()) {
            throw QString("No such directory: %1").arg(dir.absolutePath());
        }

        if (images.length() == 0) {
            throw QString("No images");

        } else if (images.length() == 1) {
            name = basename + ".png";
            int i = 1;
            while (dir.exists(name)) {
                name = QString("%1(%2).png").arg(basename).arg(i++);
            }

            QString filePath = dir.absoluteFilePath(name);
            QFile file(filePath);
            if (!file.open(QIODevice::WriteOnly)) {
                throw QString("Failed to open file: %1").arg(filePath);
            }
            if (file.write(QByteArray::fromBase64(images[0].toLocal8Bit())) == -1) {
                throw QString("Failed to write file: %1").arg(filePath);
            }

            QFileInfo info(filePath);
            QString ret = Metadata(info).toString();

            binder->success(ret);

        } else {
            QString imagesDirName = QString("%1_images").arg(basename);

            int i = 1;
            while (dir.exists(imagesDirName)) {
                imagesDirName = QString("%1_images(%2)").arg(basename).arg(i++);
            }

            QString imagesDirPath = dir.absoluteFilePath(imagesDirName);

            if (!dir.mkdir(imagesDirName)) {
                throw QString("Failed to mkdir: %1").arg(imagesDirPath);
            }

            if (!dir.cd(imagesDirName)) {
                throw QString("Failed to cd: %1").arg(imagesDirPath);
            }

            for (i = 0; i < images.length(); i++) {
                QString filePath = dir.absoluteFilePath(QString("image%1.png").arg(i + 1, 3, 10, QChar('0')));

                QFile file(filePath);
                if (!file.open(QIODevice::WriteOnly)) {
                    throw QString("Failed to open file: %1").arg(filePath);
                }

                if (file.write(QByteArray::fromBase64(images[i].toLocal8Bit())) == -1) {
                    throw QString("Failed to write file: %1").arg(filePath);
                }
            }

            QFileInfo info(dir.path());
            QString ret = Metadata(info).toString();

            binder->success(ret);
        }

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::network(Binder *binder, QStringList &args)
{
    //qDebug() << "network";

    try {
        QString &url = args[0];
        //qDebug() << "url:" << url;
        QString &type = args[1];

        QNetworkAccessManager *manager = new QNetworkAccessManager();

        QNetworkRequest req = QNetworkRequest(QUrl(url));
        req.setOriginatingObject(binder);

        connect(manager, SIGNAL(finished(QNetworkReply*)), SLOT(network_reply(QNetworkReply *)));

        if (type == "GET") {
            manager->get(req);
        } else if (type == "POST") {
            QString &data = args[2];
            //qDebug() << "data: " << data;
            manager->post(req, data.toUtf8());
        }

    } catch (QString err) {
        binder->error(err);
    }
}


void FileManagerService::network_reply(QNetworkReply *reply)
{
    //qDebug() << "network_reply";

    Binder *binder = (Binder *)reply->request().originatingObject();

    if (reply->error() == QNetworkReply::NoError) {
        QString ret = reply->readAll();
        binder->success(ret);
    } else {
        QString err = reply->errorString();
        binder->error(err);
    }
    emit signal_network_finished(reply->manager());
}

void FileManagerService::network_finished(QNetworkAccessManager *manager)
{
    //qDebug() << "network_finished";
    delete manager;
}

void FileManagerService::openurl(QString url)
{
#if defined(Q_OS_LINUX)
    QProcess::startDetached("xdg-open", QStringList() << url);
#elif defined(Q_OS_WIN)
    QProcess::startDetached("rundll32.exe", QStringList() << "url.dll,FileProtocolHandler" << url);
#endif
}

void FileManagerService::debug(QString msg)
{
    qDebug() << "DEBUG: " << msg;
}
